/* float.c     floating-point constant support for the Netwide Assembler
 *
 * The Netwide Assembler is copyright (C) 1996 Simon Tatham and
 * Julian Hall. All rights reserved. The software is
 * redistributable under the licence given in the file "Licence"
 * distributed in the NASM archive.
 *
 * initial version 13/ix/96 by Simon Tatham
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* #define FORKED_NASM */                                  /* define to target forked NASM */

#include "nasm.h"
#include "float.h"

/*
 * --------------------
 *  external variables
 * --------------------
 */

#ifdef FORKED_NASM
extern err_t err;                                          /* global error handling routine */
#endif

/*
 * -----------------
 *  local variables
 * -----------------
 */

#ifndef FORKED_NASM
static efunc err;                                          /* local error handling routine */
#endif

static Bool_t daz = FLOAT_DAZ_DEFAULT;                     /* denormals as zero */

static Uint8_t rc = FLOAT_RC_DEFAULT;                      /* rounding control */

/*
 * -----------
 *  constants
 * -----------
 */

#define MANT_WORDS 6                                       /* 64 bits + 32 for accuracy = 96 */
#define MANT_DIGITS 28                                     /* 29 digits don't fit in 96 bits */

/*
 * ---------------------------------------------------------------------------
 *  multiply
 * ---------------------------------------------------------------------------
 */

static Sint32_t multiply ( Uint16_t *to,
                           Uint16_t *from )
{
  Uint32_t temp[MANT_WORDS * 2];
  Sint32_t i, j;

  /* guaranteed that top bit of 'from' is set -- so we only have to worry about _one_ bit shift to the left */

  for ( i = 0; i < MANT_WORDS * 2; i++ ) {
    temp[i] = 0;
  }
  for ( i = 0; i < MANT_WORDS; i++ ) {
    for ( j = 0; j < MANT_WORDS; j++ ) {
      Uint32_t n;

      n = ( Uint32_t ) to[i] * ( Uint32_t ) from[j];
      temp[i + j] += n >> 16;
      temp[i + j + 1] += n & 0xFFFF;
    }
  }
  for ( i = MANT_WORDS * 2; --i; ) {
    temp[i - 1] += temp[i] >> 16;
    temp[i] &= 0xFFFF;
  }
  if ( temp[0] & 0x8000 ) {
    for ( i = 0; i < MANT_WORDS; i++ ) {
      to[i] = temp[i] & 0xFFFF;
    }
    return 0;
  } else {
    for ( i = 0; i < MANT_WORDS; i++ ) {
      to[i] = ( temp[i] << 1 ) + !!( temp[i + 1] & 0x8000 );
    }
    return -1;
  }
}

/*
 * ---------------------------------------------------------------------------
 *  convert
 * ---------------------------------------------------------------------------
 */

static Bool_t flconvert ( const Sint8_t *string,
                          Uint16_t *mant,
                          Sint32_t *exponent )
{
  Sint8_t digits[MANT_DIGITS];
  Sint8_t *p, *q, *r;
  Uint16_t mult[MANT_WORDS], bit;
  Uint16_t *m;
  Sint32_t tenpwr, twopwr;
  Sint32_t extratwos;
  Bool_t started, seendot, warned;

  p = digits;
  tenpwr = 0;
  started = seendot = warned = FALSE;
  while ( *string && *string != 'E' && *string != 'e' ) {
    if ( *string == '.' ) {
      if ( !seendot ) {
        seendot = TRUE;
      } else {
        err ( ERR_TYPE_ERROR, "too many periods in floating-point constant" );
        return FALSE;
      }
    } else if ( *string >= '0' && *string <= '9' ) {
      if ( *string == '0' && !started ) {
        if ( seendot ) {
          tenpwr--;
        }
      } else {
        started = TRUE;
        if ( p < digits + sizeof ( digits ) ) {
          *p++ = *string - '0';
        } else {
          if ( !warned ) {
            err ( ERR_TYPE_WARN | ERR_WARN_FPA, "floating-point constant significand contains more than %i digits", MANT_DIGITS );
            warned = TRUE;
          }
        }
        if ( !seendot ) {
          tenpwr++;
        }
      }
    } else if ( *string == '_' ) {
      /* do nothing */
    } else {
      err ( ERR_TYPE_ERROR, "invalid character in floating-point constant significand: '%c'", *string );
      return FALSE;
    }
    string++;
  }
  if ( *string ) {
    Sint32_t i = 0;
    Bool_t neg = FALSE;

    string++;                                              /* eat the E */
    if ( *string == '+' ) {
      string++;
    } else if ( *string == '-' ) {
      neg = TRUE;
      string++;
    }
    while ( *string ) {
      if ( *string >= '0' && *string <= '9' ) {
        i = ( i * 10 ) + ( *string - '0' );
        /*
         * To ensure that underflows and overflows are handled properly we
         * must avoid wraparounds of the signed integer value that is used
         * to hold the exponent. Therefore we cap the exponent at +/-5000,
         * which is slightly more/less than what's required for normal and
         * denormal numbers in single, double, and extended precision, but
         * sufficient to avoid signed integer wraparound.
         */
        if ( i > 5000 ) {
          break;
        }
      } else if ( *string == '_' ) {
        /* do nothing */
      } else {
        err ( ERR_TYPE_ERROR, "invalid character in floating-point constant exponent: '%c'", *string );
        return FALSE;
      }
      string++;
    }
    if ( neg ) {
      i = 0 - i;
    }
    tenpwr += i;
  }
  /*
   * At this point the memory interval [digits,p) contains a
   * series of decimal digits zzzzzzz such that our number X
   * satisfies X = 0.zzzzzzz * 10^tenpwr.
   */
  bit = 0x8000;
  for ( m = mant; m < mant + MANT_WORDS; m++ ) {
    *m = 0;
  }
  m = mant;
  q = digits;
  started = FALSE;
  twopwr = 0;
  while ( m < mant + MANT_WORDS ) {
    Uint16_t carry = 0;

    while ( p > q && !p[-1] ) {
      p--;
    }
    if ( p <= q ) {
      break;
    }
    for ( r = p; r-- > q; ) {
      Sint32_t i;

      i = 2 * *r + carry;
      if ( i >= 10 ) {
        carry = 1;
        i -= 10;
      } else {
        carry = 0;
      }
      *r = i;
    }
    if ( carry ) {
      *m |= bit;
      started = TRUE;
    }
    if ( started ) {
      if ( bit == 1 ) {
        bit = 0x8000;
        m++;
      } else {
        bit >>= 1;
      }
    } else {
      twopwr--;
    }
  }
  twopwr += tenpwr;
  /*
   * At this point the 'mant' array contains the first six
   * fractional places of a base-2^16 real number, which when
   * multiplied by 2^twopwr and 5^tenpwr gives X. So now we
   * really do multiply by 5^tenpwr.
   */
  if ( tenpwr < 0 ) {
    for ( m = mult; m < mult + MANT_WORDS; m++ ) {
      *m = 0xCCCC;
    }
    extratwos = -2;
    tenpwr = -tenpwr;
    /*
     * If tenpwr was 1000...000b, then it becomes 1000...000b. See
     * the "ANSI C" comment below for more details on that case.
     *
     * Because we already truncated tenpwr to +5000...-5000 inside
     * the exponent parsing code, this shouldn't happen though.
     */
  } else if ( tenpwr > 0 ) {
    mult[0] = 0xA000;
    for ( m = mult + 1; m < mult + MANT_WORDS; m++ ) {
      *m = 0;
    }
    extratwos = 3;
  } else {
    extratwos = 0;
  }
  while ( tenpwr ) {
    if ( tenpwr & 1 ) {
      twopwr += extratwos + multiply ( mant, mult );
    }
    extratwos = extratwos * 2 + multiply ( mult, mult );
    tenpwr >>= 1;
    /*
     * In ANSI C, the result of right-shifting a signed integer is
     * considered implementation-specific. To ensure that the loop
     * terminates even if tenpwr was 1000...000b to begin with, we
     * manually clear the MSB, in case a 1 was shifted in.
     *
     * Because we already truncated tenpwr to +5000...-5000 inside
     * the exponent parsing code, this shouldn't matter; neverthe-
     * less it is the right thing to do here.
     */
    tenpwr &= ( unsigned ) -1 >> 1;
  }
  /*
   * Conversion is done. The elements of 'mant' contain the first
   * fractional places of a base-2^16 real number in [0.5,1)
   * which we can multiply by 2^twopwr to get X. Or, of course,
   * it contains zero.
   */
  *exponent = twopwr;
  return TRUE;
}

/*
 * ---------------------------------------------------------------------------
 *  shift a mantissa to the right by i (i < 16) bits
 * ---------------------------------------------------------------------------
 */

static void shr ( Uint16_t *mant,
                  Sint32_t i )
{
  Uint16_t n = 0, m;
  Sint32_t j;

  for ( j = 0; j < MANT_WORDS; j++ ) {
    m = ( mant[j] << ( 16 - i ) ) & 0xFFFF;
    mant[j] = ( mant[j] >> i ) | n;
    n = m;
  }
}

/*
 * ---------------------------------------------------------------------------
 *  round a mantissa off after i words
 * ---------------------------------------------------------------------------
 */

#define ROUND_COLLECT_BITS                                 \
  for ( j = i; j < MANT_WORDS; j++ ) {                     \
    m = m | mant[j];                                       \
  }

#define ROUND_ABS_DOWN                                     \
  for ( j = i; j < MANT_WORDS; j++ ) {                     \
    mant[j] = 0x0000;                                      \
  }

#define ROUND_ABS_UP                                       \
  do {                                                     \
    ++mant[--i];                                           \
    mant[i] &= 0xFFFF;                                     \
  } while ( i > 0 && !mant[i] );                           \
  return !i && !mant[i];

static Sint32_t round ( Sint32_t sign,
                        Uint16_t *mant,
                        Sint32_t i )
{
  Uint16_t m = 0;
  Sint32_t j;

  if ( ( sign == 0x0000 ) || ( sign == 0x8000 ) ) {
    if ( rc == FLOAT_RC_NEAR ) {
      if ( mant[i] & 0x8000 ) {
        ROUND_ABS_UP;
      }
    } else if ( ( ( sign == 0x0000 ) && ( rc == FLOAT_RC_DOWN ) ) || ( ( sign == 0x8000 ) && ( rc == FLOAT_RC_UP ) ) ) {
      ROUND_COLLECT_BITS;
      if ( m ) {
        ROUND_ABS_DOWN;
      }
    } else if ( ( ( sign == 0x0000 ) && ( rc == FLOAT_RC_UP ) ) || ( ( sign == 0x8000 ) && ( rc == FLOAT_RC_DOWN ) ) ) {
      ROUND_COLLECT_BITS;
      if ( m ) {
        ROUND_ABS_UP;
      }
    } else if ( rc == FLOAT_RC_ZERO ) {
      ROUND_ABS_DOWN;
    } else {
      err ( ERR_TYPE_PANIC, "round() can't handle rc=%i", rc );
    }
  } else {
    err ( ERR_TYPE_PANIC, "round() can't handle sign=%i", sign );
  }
  return 0;
}

/*
 * ---------------------------------------------------------------------------
 *  put a mantissa word into two little-endian bytes
 * ---------------------------------------------------------------------------
 */

#define PUT(a,b) ( ( (a)[0] = (b) & 0xFF ), ( (a)[1] = ( (b) >> 8 ) & 0xFF ) )

/*
 * ---------------------------------------------------------------------------
 *  emit a printf()-like debug message... but only if DEBUG_FLOAT was defined
 * ---------------------------------------------------------------------------
 */

#ifdef DEBUG_FLOAT
#define dprintf printf
#else
static void null_printf ( const Sint8_t *format, ... )
{
  ( void ) format;
}
#define dprintf null_printf
#endif

/*
 * ---------------------------------------------------------------------------
 *  convert to REAL4
 * ---------------------------------------------------------------------------
 */

static Bool_t to_real4 ( const Sint8_t *string,
                         Sint32_t sign,
                         Uint8_t *result )
{
  Uint16_t mant[MANT_WORDS];
  Sint32_t exponent;

  sign = ( sign < 0 ? 0x8000L : 0L );
  dprintf ( "DD = %s\n", string );
  if ( !flconvert ( string, mant, &exponent ) ) {
    dprintf ( "invalid\n");
    PUT ( result + 2, 0x0000 | sign );
    PUT ( result + 0, 0x0000 );
    dprintf ( "\n" );
    return FALSE;
  }
  dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
  if ( mant[0] & 0x8000 ) {
    exponent--;
    exponent += 0x7F;
    shr ( mant, 8 );
    dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after exponent+127 and mant>>8\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
    round ( sign, mant, 2 );
    dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after round=2)\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
    if ( mant[0] & 0x0100 ) {
      shr ( mant, 1 );
      exponent++;
      dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after exponent++ and mant>>1)\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
    }
    if ( exponent > ( 0xFF - 1 ) ) {
      dprintf ( "overflow\n" );
      err ( ERR_TYPE_WARN | ERR_WARN_FPO, "overflow in floating-point constant resulted in %cINFINITY", sign == 0 ? '+' : '-' );
      PUT ( result + 2, 0x7F80 | sign );
      PUT ( result + 0, 0x0000 );
    } else if ( exponent > 0 ) {
      dprintf ( "normal\n" );
      mant[0] &= 0x007F;
      dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after leading 1 removed)\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
      PUT ( result + 2, ( exponent << 7 ) | mant[0] | sign );
      PUT ( result + 0, mant[1] );
    } else if ( exponent >= ( 1 - 23 ) ) {
      if ( daz == FLOAT_DAZ_OFF ) {
        dprintf ( "denormal\n" );
        err ( ERR_TYPE_WARN | ERR_WARN_FPD, "tiny floating-point constant resulted in denormal number" );
        while ( exponent < 1 ) {
          shr ( mant, 1 );
          exponent++;
          dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after mant>>1)\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
        }
        PUT ( result + 2, ( 0 << 7 ) | mant[0] | sign );
        PUT ( result + 0, mant[1] );
      } else {
        dprintf ( "daz\n" );
        err ( ERR_TYPE_WARN | ERR_WARN_FPU, "tiny floating-point constant resulted in %c0.0 due to DAZ", sign == 0 ? '+' : '-' );
        PUT ( result + 2, 0x0000 | sign );
        PUT ( result + 0, 0x0000 );
      }
    } else {
      dprintf ( "underflow\n" );
      err ( ERR_TYPE_WARN | ERR_WARN_FPU, "underflow in floating-point constant resulted in %c0.0", sign == 0 ? '+' : '-' );
      PUT ( result + 2, 0x0000 | sign );
      PUT ( result + 0, 0x0000 );
    }
  } else {
    dprintf ( "zero\n" );
    PUT ( result + 2, 0x0000 | sign );
    PUT ( result + 0, 0x0000 );
  }
  dprintf ( "\n" );
  return TRUE;
}

/*
 * ---------------------------------------------------------------------------
 *  convert to REAL8
 * ---------------------------------------------------------------------------
 */

static Bool_t to_real8 ( const Sint8_t *string,
                         Sint32_t sign,
                         Uint8_t *result )
{
  Uint16_t mant[MANT_WORDS];
  Sint32_t exponent;

  sign = ( sign < 0 ? 0x8000L : 0L );
  dprintf ( "DQ = %s\n", string );
  if ( !flconvert ( string, mant, &exponent ) ) {
    dprintf ( "invalid\n");
    PUT ( result + 6, 0x0000 | sign );
    PUT ( result + 4, 0x0000 );
    PUT ( result + 2, 0x0000 );
    PUT ( result + 0, 0x0000 );
    dprintf ( "\n" );
    return FALSE;
  }
  dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
  if ( mant[0] & 0x8000 ) {
    exponent--;
    exponent += 0x3FF;
    shr ( mant, 11 );
    dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after exponent+1023 and mant>>11\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
    round ( sign, mant, 4 );
    dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after round=4)\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
    if ( mant[0] & 0x0020 ) {
      shr ( mant, 1 );
      exponent++;
      dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after exponent++ and mant>>1)\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
    }
    if ( exponent > ( 0x7FF - 1 ) ) {
      dprintf ( "overflow\n" );
      err ( ERR_TYPE_WARN | ERR_WARN_FPO, "overflow in floating-point constant resulted in %cINFINITY", sign == 0 ? '+' : '-' );
      PUT ( result + 6, 0x7FF0 | sign );
      PUT ( result + 4, 0x0000 );
      PUT ( result + 2, 0x0000 );
      PUT ( result + 0, 0x0000 );
    } else if ( exponent > 0 ) {
      dprintf ( "normal\n" );
      mant[0] &= 0x000F;
      dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after leading 1 removed)\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
      PUT ( result + 6, ( exponent << 4 ) | mant[0] | sign );
      PUT ( result + 4, mant[1] );
      PUT ( result + 2, mant[2] );
      PUT ( result + 0, mant[3] );
    } else if ( exponent >= ( 1 - 52 ) ) {
      if ( daz == FLOAT_DAZ_OFF ) {
        dprintf ( "denormal\n" );
        err ( ERR_TYPE_WARN | ERR_WARN_FPD, "tiny floating-point constant resulted in denormal number" );
        while ( exponent < 1 ) {
          shr ( mant, 1 );
          exponent++;
          dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after mant>>1)\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
        }
        PUT ( result + 6, ( 0 << 4 ) | mant[0] | sign );
        PUT ( result + 4, mant[1] );
        PUT ( result + 2, mant[2] );
        PUT ( result + 0, mant[3] );
      } else {
        dprintf ( "daz\n" );
        err ( ERR_TYPE_WARN | ERR_WARN_FPU, "tiny floating-point constant resulted in %c0.0 due to DAZ", sign == 0 ? '+' : '-' );
        PUT ( result + 6, 0x0000 | sign );
        PUT ( result + 4, 0x0000 );
        PUT ( result + 2, 0x0000 );
        PUT ( result + 0, 0x0000 );
      }
    } else {
      dprintf ( "underflow\n" );
      err ( ERR_TYPE_WARN | ERR_WARN_FPU, "underflow in floating-point constant resulted in %c0.0", sign == 0 ? '+' : '-' );
      PUT ( result + 6, 0x0000 | sign );
      PUT ( result + 4, 0x0000 );
      PUT ( result + 2, 0x0000 );
      PUT ( result + 0, 0x0000 );
    }
  } else {
    dprintf ( "zero\n" );
    PUT ( result + 6, 0x0000 | sign );
    PUT ( result + 4, 0x0000 );
    PUT ( result + 2, 0x0000 );
    PUT ( result + 0, 0x0000 );
  }
  dprintf ( "\n" );
  return TRUE;
}

/*
 * ---------------------------------------------------------------------------
 *  convert to REAL10
 * ---------------------------------------------------------------------------
 */

static Bool_t to_real10 ( const Sint8_t *string,
                          Sint32_t sign,
                          Uint8_t *result )
{
  Uint16_t mant[MANT_WORDS];
  Sint32_t exponent;

  sign = ( sign < 0 ? 0x8000L : 0L );
  dprintf ( "DT = %s\n", string );
  if ( !flconvert ( string, mant, &exponent ) ) {
    dprintf ( "invalid\n");
    PUT ( result + 8, 0x0000 | sign );
    PUT ( result + 6, 0x0000 );
    PUT ( result + 4, 0x0000 );
    PUT ( result + 2, 0x0000 );
    PUT ( result + 0, 0x0000 );
    dprintf ( "\n" );
    return FALSE;
  }
  dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
  if ( mant[0] & 0x8000 ) {
    exponent--;
    exponent += 0x3FFF;
    dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after exponent+16383\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
    if ( round ( sign, mant, 4 ) ) {
      dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after round=4)\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
      shr ( mant, 1 );
      exponent++;
      mant[0] |= 0x8000;
      dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (exponent++ and mant>>1)\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
    } else {
      dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after round=4)\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
    }
    if ( exponent > ( 0x7FFF - 1 ) ) {
      dprintf ( "overflow\n" );
      err ( ERR_TYPE_WARN | ERR_WARN_FPO, "overflow in floating-point constant resulted in %cINFINITY", sign == 0 ? '+' : '-' );
      PUT ( result + 8, 0x7FFF | sign );
      PUT ( result + 6, 0x8000 );
      PUT ( result + 4, 0x0000 );
      PUT ( result + 2, 0x0000 );
      PUT ( result + 0, 0x0000 );
    } else if ( exponent > 0 ) {
      dprintf ( "normal\n" );
      PUT ( result + 8, exponent << 0 | sign );
      PUT ( result + 6, mant[0] );
      PUT ( result + 4, mant[1] );
      PUT ( result + 2, mant[2] );
      PUT ( result + 0, mant[3] );
    } else if ( exponent >= ( 1 - 63 ) ) {
      if ( daz == FLOAT_DAZ_OFF ) {
        dprintf ( "denormal\n" );
        err ( ERR_TYPE_WARN | ERR_WARN_FPD, "tiny floating-point constant resulted in denormal number" );
        while ( exponent < 1 ) {
          shr ( mant, 1 );
          exponent++;
          dprintf ( "mant=%04X_%04X_%04X_%04X_%04X_%04X * 2^%i (after mant>>1)\n", mant[0],mant[1],mant[2],mant[3],mant[4],mant[5], exponent );
        }
        PUT ( result + 8, 0 << 0 | sign );
        PUT ( result + 6, mant[0] );
        PUT ( result + 4, mant[1] );
        PUT ( result + 2, mant[2] );
        PUT ( result + 0, mant[3] );
      } else {
        dprintf ( "daz\n" );
        err ( ERR_TYPE_WARN | ERR_WARN_FPU, "tiny floating-point constant resulted in %c0.0 due to DAZ", sign == 0 ? '+' : '-' );
        PUT ( result + 8, 0x0000 | sign );
        PUT ( result + 6, 0x0000 );
        PUT ( result + 4, 0x0000 );
        PUT ( result + 2, 0x0000 );
        PUT ( result + 0, 0x0000 );
      }
    } else {
      dprintf ( "underflow\n" );
      err ( ERR_TYPE_WARN | ERR_WARN_FPU, "underflow in floating-point constant resulted in %c0.0", sign == 0 ? '+' : '-' );
      PUT ( result + 8, 0x0000 | sign );
      PUT ( result + 6, 0x0000 );
      PUT ( result + 4, 0x0000 );
      PUT ( result + 2, 0x0000 );
      PUT ( result + 0, 0x0000 );
    }
  } else {
    dprintf ( "zero\n" );
    PUT ( result + 8, 0x0000 | sign );
    PUT ( result + 6, 0x0000 );
    PUT ( result + 4, 0x0000 );
    PUT ( result + 2, 0x0000 );
    PUT ( result + 0, 0x0000 );
  }
  dprintf ( "\n" );
  return TRUE;
}

/*
 * ---------------------------------------------------------------------------
 *  convert to REALn
 * ---------------------------------------------------------------------------
 */

#ifndef FORKED_NASM
Bool_t float_const ( const Sint8_t *string,
                     Sint32_t sign,
                     Uint8_t *result,
                     Sint32_t bytes,
                     efunc error )
#else
Bool_t float_const ( const Sint8_t *string,
                     Sint32_t sign,
                     Uint8_t *result,
                     Sint32_t bytes )
#endif
{
#ifndef FORKED_NASM
  err = error;
#endif
  if ( bytes == 4 ) {
    return to_real4 ( string, sign, result );
  } else if ( bytes == 8 ) {
    return to_real8 ( string, sign, result );
  } else if ( bytes == 10 ) {
    return to_real10 ( string, sign, result );
  } else {
    err ( ERR_TYPE_PANIC, "float_const() can't handle REAL%d", bytes );
    return FALSE;
  }
}

/*
 * ---------------------------------------------------------------------------
 *  query current DAZ behavior
 * ---------------------------------------------------------------------------
 */

Bool_t float_get_daz ( void )
{
  return daz;
}

/*
 * ---------------------------------------------------------------------------
 *  select desired DAZ behavior
 * ---------------------------------------------------------------------------
 */

void float_set_daz ( Bool_t state )
{
  if ( ( state == FLOAT_DAZ_OFF ) || ( state == FLOAT_DAZ_ON ) ) {
    daz = state;
    dprintf ( "set DAZ to %s\n", state == FLOAT_DAZ_OFF ? "off" : "on" );
#ifdef FORKED_NASM
  } else {
    err ( ERR_TYPE_PANIC, "float_daz() can't handle state=%i", state );
#endif
  }
}

/*
 * ---------------------------------------------------------------------------
 *  query current rounding control
 * ---------------------------------------------------------------------------
 */

Uint8_t float_get_rc ( void )
{
  return rc;
}

/*
 * ---------------------------------------------------------------------------
 *  select desired rounding control
 * ---------------------------------------------------------------------------
 */

void float_set_rc ( Uint8_t type )
{
  if ( ( type == FLOAT_RC_NEAR ) || ( type == FLOAT_RC_DOWN ) || ( type == FLOAT_RC_UP ) || ( type == FLOAT_RC_ZERO ) ) {
    rc = type;
    dprintf ( "set RC to %s\n", type == FLOAT_RC_NEAR ? "near" : type == FLOAT_RC_DOWN ? "down" : type == FLOAT_RC_UP ? "up" : "zero" );
#ifdef FORKED_NASM
  } else {
    err ( ERR_TYPE_PANIC, "float_rc() can't handle type=%i", type );
#endif
  }
}

/* EOF */
